#pragma once

#define EIGEN_DONT_ALIGN_STATICALLY
#include <Eigen/Dense>

#define GLM_FORCE_RADIANS
#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtx/string_cast.hpp>

#include "Collider.h"
#include "Plane.h"

#include "Shape.h"
#include "MatrixStack.h"
#include "Program.h"

#ifndef M_PI
    #define M_PI 3.14159265358979323846
#endif

class Box : public Collider
{
public:
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW

    Eigen::Vector3d position;
    Eigen::Vector3d rotation;
    Eigen::Vector3d scale;

    glm::mat4 rotMat;
    std::shared_ptr<Plane> plane;
    std::shared_ptr<Shape> boxShape;

    Box(Eigen::Vector3d pos, Eigen::Vector3d rot, Eigen::Vector3d _scale){
        position = pos;
        rotation = rot;
        scale = _scale;

        plane = std::make_shared<Plane>();
    }

    void update() {
        rotMat = glm::rotate(glm::rotate(glm::rotate(glm::mat4(1.0f), 
            (float) (rotation(0)*M_PI/180), glm::vec3(1, 0, 0)), 
            (float) (rotation(1)*M_PI/180), glm::vec3(0, 1, 0)), 
            (float) (rotation(2)*M_PI/180), glm::vec3(0, 0, 1));
        glm::vec3 xV = glm::vec3(rotMat * glm::vec4(1, 0, 0, 0));
        glm::vec3 yV = glm::vec3(rotMat * glm::vec4(0, 1, 0, 0));
        glm::vec3 zV = glm::vec3(rotMat * glm::vec4(0, 0, 1, 0));

        plane->normal = Eigen::Vector3d(yV.x, yV.y, yV.z).normalized();
        plane->center = position + plane->normal * scale(1);

        plane->xVec = Eigen::Vector3d(xV.x, xV.y, xV.z).normalized();
        plane->zVec = Eigen::Vector3d(zV.x, zV.y, zV.z).normalized();
        
        plane->xSize = scale(0);
        plane->zSize = scale(2);
    }

	double intersect(Eigen::Vector3d point, Eigen::Vector3d dir) {
        return plane->intersect(point, dir);
    }

    void load(const std::string &RESOURCE_DIR) {
        boxShape = std::make_shared<Shape>();
	    boxShape->loadMesh(RESOURCE_DIR + "cube.obj");
    }

    void init() {
        boxShape->init();
    }

    void draw(std::shared_ptr<MatrixStack> MV, const std::shared_ptr<Program> p) const {
        
        MV->pushMatrix();

        MV->translate(position(0), position(1), position(2));
        MV->multMatrix(rotMat);

        MV->scale(glm::vec3(scale(0), scale(1), scale(2)));

        // MV->scale(2.0/30.0);
        // MV->rotate(M_PI/2, 0, 1, 0);
        // MV->translate(0, 0, 2.75);
        glUniformMatrix4fv(p->getUniform("MV"), 1, GL_FALSE, glm::value_ptr(MV->topMatrix()));
        // glUniform1f(p->getUniform("alpha"),  0.8);
        glUniform3fv(p->getUniform("kdFront"), 1, Eigen::Vector3f(0.5, 0.5, 0.5).data());
        // glUniform3fv(p->getUniform("kdBack"),  1, Eigen::Vector3f(0.1, 0.1, 0.1).data());
        boxShape->draw(p);

        MV->popMatrix();

    }
};