#include <iostream>

#define GLM_FORCE_RADIANS
#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>

#include "Cloth.h"
#include "Particle.h"
#include "Spring.h"
#include "MatrixStack.h"
#include "Program.h"
#include "GLSL.h"

using namespace std;
using namespace Eigen;

typedef Eigen::Triplet<double> Td;

shared_ptr<Spring> createSpring(const shared_ptr<Particle> p0, const shared_ptr<Particle> p1, double E)
{
	auto s = make_shared<Spring>(p0, p1);
	s->E = E;
	Vector3d x0 = p0->x;
	Vector3d x1 = p1->x;
	Vector3d dx = x1 - x0;
	s->L = dx.norm();
	return s;
}

Cloth::Cloth(int rows, int cols,
			 const Vector3d &x00,
			 const Vector3d &x01,
			 const Vector3d &x10,
			 const Vector3d &x11,
			 double mass,
			 double stiffness)
{
	assert(rows > 1);
	assert(cols > 1);
	assert(mass > 0.0);
	assert(stiffness > 0.0);
	
	this->rows = rows;
	this->cols = cols;
	
	// Create particles
	n = 0;
	double r = 0.02; // Used for collisions
	int nVerts = rows*cols;
	for(int i = 0; i < rows; ++i) {
		double u = i / (rows - 1.0);
		Vector3d x0 = (1 - u)*x00 + u*x10;
		Vector3d x1 = (1 - u)*x01 + u*x11;
		for(int j = 0; j < cols; ++j) {
			double v = j / (cols - 1.0);
			Vector3d x = (1 - v)*x0 + v*x1;
			auto p = make_shared<Particle>();
			particles.push_back(p);
			p->r = r;
			p->x = x;
			p->v << 0.0, 0.0, 0.0;
			p->m = mass/(nVerts);
			// Pin two particles
			if(i == 0 && (j == 0 || j == cols-1)) {
				p->fixed = true;
				p->i = -1;
			} else {
				p->fixed = false;
				p->i = n;
				n += 3;
			}
		}
	}
	
	// Create x springs
	for(int i = 0; i < rows; ++i) {
		for(int j = 0; j < cols-1; ++j) {
			int k0 = i*cols + j;
			int k1 = k0 + 1;
			springs.push_back(createSpring(particles[k0], particles[k1], stiffness));
		}
	}
	
	// Create y springs
	for(int j = 0; j < cols; ++j) {
		for(int i = 0; i < rows-1; ++i) {
			int k0 = i*cols + j;
			int k1 = k0 + cols;
			springs.push_back(createSpring(particles[k0], particles[k1], stiffness));
		}
	}
	
	// Create shear springs
	for(int i = 0; i < rows-1; ++i) {
		for(int j = 0; j < cols-1; ++j) {
			int k00 = i*cols + j;
			int k10 = k00 + 1;
			int k01 = k00 + cols;
			int k11 = k01 + 1;
			springs.push_back(createSpring(particles[k00], particles[k11], stiffness));
			springs.push_back(createSpring(particles[k10], particles[k01], stiffness));
		}
	}
	
	// Create x bending springs
	for(int i = 0; i < rows; ++i) {
		for(int j = 0; j < cols-2; ++j) {
			int k0 = i*cols + j;
			int k2 = k0 + 2;
			springs.push_back(createSpring(particles[k0], particles[k2], stiffness));
		}
	}
	
	// Create y bending springs
	for(int j = 0; j < cols; ++j) {
		for(int i = 0; i < rows-2; ++i) {
			int k0 = i*cols + j;
			int k2 = k0 + 2*cols;
			springs.push_back(createSpring(particles[k0], particles[k2], stiffness));
		}
	}

	// Build system matrices and vectors
	// M.resize(n,n);
	// K.resize(n,n);
	v.resize(n);
	f.resize(n);
	
	// Build vertex buffers
	posBuf.clear();
	norBuf.clear();
	// texBuf.clear();
	eleBuf.clear();
	posBuf.resize(nVerts*3);
	norBuf.resize(nVerts*3);
	updatePosNor();

	// // Texture coordinates (don't change)
	// for(int i = 0; i < rows; ++i) {
	// 	for(int j = 0; j < cols; ++j) {
	// 		texBuf.push_back(i/(rows-1.0));
	// 		texBuf.push_back(j/(cols-1.0));
	// 	}
	// }

	// Elements (don't change)
	for(int i = 0; i < rows-1; ++i) {
		for(int j = 0; j < cols; ++j) {
			int k0 = i*cols + j;
			int k1 = k0 + cols;
			// Triangle strip
			eleBuf.push_back(k0);
			eleBuf.push_back(k1);
		}
	}
}

Cloth::~Cloth()
{
}

void Cloth::tare()
{
	for(int k = 0; k < (int)particles.size(); ++k) {
		particles[k]->tare();
	}
}

void Cloth::reset()
{
	for(int k = 0; k < (int)particles.size(); ++k) {
		particles[k]->reset();
	}
	updatePosNor();
}

void Cloth::updatePosNor()
{
	// Position
	for(int i = 0; i < rows; ++i) {
		for(int j = 0; j < cols; ++j) {
			int k = i*cols + j;
			Vector3d x = particles[k]->x;
			posBuf[3*k+0] = x(0);
			posBuf[3*k+1] = x(1);
			posBuf[3*k+2] = x(2);
		}
	}
	
	// Normal
	for(int i = 0; i < rows; ++i) {
		for(int j = 0; j < cols; ++j) {
			// Each particle has four neighbors
			//
			//      v1
			//     /|\
			// u0 /_|_\ u1
			//    \ | /
			//     \|/
			//      v0
			//
			// Use these four triangles to compute the normal
			int k = i*cols + j;
			int ku0 = k - 1;
			int ku1 = k + 1;
			int kv0 = k - cols;
			int kv1 = k + cols;
			Vector3d x = particles[k]->x;
			Vector3d xu0, xu1, xv0, xv1, dx0, dx1, c;
			Vector3d nor(0.0, 0.0, 0.0);
			int count = 0;
			// Top-right triangle
			if(j != cols-1 && i != rows-1) {
				xu1 = particles[ku1]->x;
				xv1 = particles[kv1]->x;
				dx0 = xu1 - x;
				dx1 = xv1 - x;
				c = dx0.cross(dx1);
				nor += c.normalized();
				++count;
			}
			// Top-left triangle
			if(j != 0 && i != rows-1) {
				xu1 = particles[kv1]->x;
				xv1 = particles[ku0]->x;
				dx0 = xu1 - x;
				dx1 = xv1 - x;
				c = dx0.cross(dx1);
				nor += c.normalized();
				++count;
			}
			// Bottom-left triangle
			if(j != 0 && i != 0) {
				xu1 = particles[ku0]->x;
				xv1 = particles[kv0]->x;
				dx0 = xu1 - x;
				dx1 = xv1 - x;
				c = dx0.cross(dx1);
				nor += c.normalized();
				++count;
			}
			// Bottom-right triangle
			if(j != cols-1 && i != 0) {
				xu1 = particles[kv0]->x;
				xv1 = particles[ku1]->x;
				dx0 = xu1 - x;
				dx1 = xv1 - x;
				c = dx0.cross(dx1);
				nor += c.normalized();
				++count;
			}
			nor /= count;
			nor.normalize();
			norBuf[3*k+0] = nor(0);
			norBuf[3*k+1] = nor(1);
			norBuf[3*k+2] = nor(2);
		}
	}
}

template <typename Derived>
void addMatToTriplets(const DenseBase<Derived> &mat, vector<Td> &triplets, int iOff, int jOff){
	for(int i = 0; i < mat.rows(); i++){
		for(int j = 0; j < mat.rows(); j++){
			double val = mat(i,j);
			if(val != 0){
				triplets.push_back(Td(i + iOff, j + jOff, val));
			}
		}
	}
}

void Cloth::step(double h, const Vector3d &grav, const vector< shared_ptr<Particle> > spheres)
{
	// M.setZero();
	// K.setZero();
	v.setZero();
	f.setZero();

	//
	// IMPLEMENT ME!
	//

	vector<Td> Mtrip;
	vector<Td> Ktrip;

	for(shared_ptr<Particle> particle : particles){
		if(!particle->fixed){
			// Matrix3d m;
			// m << particles[i]->m, 0, 0, 
			// 	0, particles[i]->m, 0, 
			// 	0, 0, particles[i]->m;

			// M.block<3,3>(particles[i]->i, particles[i]->i) = m;
			for(int i = 0; i < 3; i++)
				Mtrip.push_back(Td(particle->i + i, particle->i + i, particle->m));

			v.segment(particle->i, 3) = particle->v;
			f.segment(particle->i, 3) = particle->m * grav; // f = mg
		}
	}

	for(shared_ptr<Spring> spring : springs){
		Vector3d dx = spring->p1->x - spring->p0->x;
		double l = dx.norm();
		Vector3d force = spring->E * (l - spring->L) * dx / l;
		
		if(!spring->p0->fixed)
			f.segment(spring->p0->i, 3) += force;
		if(!spring->p1->fixed)
			f.segment(spring->p1->i, 3) -= force;

		Matrix3d a = (1 - (l - spring->L)/l) * (dx * dx.transpose());
		Matrix3d b = (l - spring->L)/l * (dx.transpose() * dx).value() * Matrix3d::Identity();
		Matrix3d ks = spring->E / (l * l) * (a + b);

		// if(!spring->p0->fixed)
		// 	K.block<3,3>(spring->p0->i, spring->p0->i) += -ks;
		// if(!spring->p1->fixed)
		// 	K.block<3,3>(spring->p1->i, spring->p1->i) += -ks;
		// if(!spring->p0->fixed && !spring->p1->fixed){
		// 	K.block<3,3>(spring->p0->i, spring->p1->i) += ks;
		// 	K.block<3,3>(spring->p1->i, spring->p0->i) += ks;
		// }

		if(!spring->p0->fixed)
			addMatToTriplets(-ks, Ktrip, spring->p0->i, spring->p0->i);
		if(!spring->p1->fixed)
			addMatToTriplets(-ks, Ktrip, spring->p1->i, spring->p1->i);
		if(!spring->p0->fixed && !spring->p1->fixed){
			addMatToTriplets(ks, Ktrip, spring->p0->i, spring->p1->i);
			addMatToTriplets(ks, Ktrip, spring->p1->i, spring->p0->i);
		}
	}

	// COLLISIONS

	double c = 20;

	for(shared_ptr<Particle> particle : particles){
		for(shared_ptr<Particle> sphere : spheres){
			Vector3d dx = particle->x - sphere->x;
			double l = dx.norm();
			double d = particle->r + sphere->r - l;
			
			if(d > 0){ // collision
				Vector3d collisionForce = c * d * (dx / l);
				Matrix3d stiffness = c * d * Matrix3d::Identity();

				f.segment(particle->i, 3) += collisionForce;
				// K.block<3,3>(particle->i, particle->i) += stiffness;
				addMatToTriplets(stiffness, Ktrip, particle->i, particle->i);
			}
		}
	}

	// SOLVE

	// MatrixXd A = M - h * h * K;
	// VectorXd b = M * v + h * f;

	// VectorXd newVel = A.ldlt().solve(b);
	SparseMatrix<double> M(n, n);
	M.setFromTriplets(Mtrip.begin(), Mtrip.end());

	SparseMatrix<double> K(n, n);
	K.setFromTriplets(Ktrip.begin(), Ktrip.end());

	SparseMatrix<double> A = M - h * h * K;
	VectorXd b = M * v + h * f;

	ConjugateGradient< SparseMatrix<double> > cg;
	cg.setMaxIterations(25);
	cg.setTolerance(1e-6);
	cg.compute(A);
	VectorXd newVel = cg.solveWithGuess(b, v);

	for(int i = 0; i < particles.size(); i++){
		if(!particles[i]->fixed){
			particles[i]->v = newVel.segment(particles[i]->i, 3);
		}
	}

	for(int i = 0; i < particles.size(); i++){
		if(!particles[i]->fixed){
			particles[i]->x += h * particles[i]->v;
		}
	}

	// cout << "\n---" << endl;
	// cout << A << endl;
	// cout << b << endl;
	// cout << newVel << endl;
	
	// Update position and normal buffers
	updatePosNor();
}

void Cloth::init()
{
	glGenBuffers(1, &posBufID);
	glBindBuffer(GL_ARRAY_BUFFER, posBufID);
	glBufferData(GL_ARRAY_BUFFER, posBuf.size()*sizeof(float), &posBuf[0], GL_DYNAMIC_DRAW);
	
	glGenBuffers(1, &norBufID);
	glBindBuffer(GL_ARRAY_BUFFER, norBufID);
	glBufferData(GL_ARRAY_BUFFER, norBuf.size()*sizeof(float), &norBuf[0], GL_DYNAMIC_DRAW);
	
	// glGenBuffers(1, &texBufID);
	// glBindBuffer(GL_ARRAY_BUFFER, texBufID);
	// glBufferData(GL_ARRAY_BUFFER, texBuf.size()*sizeof(float), &texBuf[0], GL_STATIC_DRAW);
	
	glGenBuffers(1, &eleBufID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, eleBufID);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, eleBuf.size()*sizeof(unsigned int), &eleBuf[0], GL_STATIC_DRAW);
	
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	
	assert(glGetError() == GL_NO_ERROR);
}

void Cloth::draw(shared_ptr<MatrixStack> MV, const shared_ptr<Program> p) const
{
	// Draw mesh
	glUniform3fv(p->getUniform("kdFront"), 1, Vector3f(1.0, 0.0, 0.0).data());
	glUniform3fv(p->getUniform("kdBack"),  1, Vector3f(1.0, 1.0, 0.0).data());
	glUniform1f(p->getUniform("alpha"),  0.8);
	MV->pushMatrix();
	glUniformMatrix4fv(p->getUniform("MV"), 1, GL_FALSE, glm::value_ptr(MV->topMatrix()));
	int h_pos = p->getAttribute("aPos");
	glEnableVertexAttribArray(h_pos);
	glBindBuffer(GL_ARRAY_BUFFER, posBufID);
	glBufferData(GL_ARRAY_BUFFER, posBuf.size()*sizeof(float), &posBuf[0], GL_DYNAMIC_DRAW);
	glVertexAttribPointer(h_pos, 3, GL_FLOAT, GL_FALSE, 0, (const void *)0);
	int h_nor = p->getAttribute("aNor");
	glEnableVertexAttribArray(h_nor);
	glBindBuffer(GL_ARRAY_BUFFER, norBufID);
	glBufferData(GL_ARRAY_BUFFER, norBuf.size()*sizeof(float), &norBuf[0], GL_DYNAMIC_DRAW);
	glVertexAttribPointer(h_nor, 3, GL_FLOAT, GL_FALSE, 0, (const void *)0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, eleBufID);
	for(int i = 0; i < rows; ++i) {
		glDrawElements(GL_TRIANGLE_STRIP, 2*cols, GL_UNSIGNED_INT, (const void *)(2*cols*i*sizeof(unsigned int)));
	}
	glDisableVertexAttribArray(h_nor);
	glDisableVertexAttribArray(h_pos);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);



	glColor3f(1.0f, 0.0f, 0.0f);

	glLineWidth(2.0f);
	glBegin(GL_LINES);

	for(shared_ptr<Spring> spring : springs) {
		glVertex3f(spring->p0->x(0), spring->p0->x(1), spring->p0->x(2));
		glVertex3f(spring->p1->x(0), spring->p1->x(1), spring->p1->x(2));
	}

	glEnd();



	MV->popMatrix();
}
